import csv
import os
from typing import Set

from pyhpo.annotations import DiseaseDict
from pyhpo.annotations import Decipher, Omim, Orpha
from pyhpo.annotations import DecipherDisease, OmimDisease, OrphaDisease
from pyhpo.parser.generics import id_from_string, remove_outcommented_rows
import pyhpo


FILENAME = "phenotype.hpoa"
# Cloumn mappings in phenotype_to_genes file
DISEASE_ID = 0
DISEASE_NAME = 1
QUALIFIER = 2
HPO_ID = 3


def _parse_phenotype_hpoa_file(path: str) -> None:
    Omim.clear()
    Orpha.clear()
    Decipher.clear()
    filename = os.path.join(path, FILENAME)
    with open(filename) as fh:
        reader = csv.reader(
            remove_outcommented_rows(
                remove_outcommented_rows(fh), ignorechar="database_id"
            ),
            delimiter="\t",
        )
        for cols in reader:
            disease_class: DiseaseDict = DiseaseDict()
            phenotype_source, phenotype_id = cols[DISEASE_ID].split(":")
            if phenotype_source == "OMIM":
                disease_class = Omim
            elif phenotype_source == "ORPHA":
                disease_class = Orpha
            elif phenotype_source == "DECIPHER":
                disease_class = Decipher
            else:
                continue

            disease = disease_class(
                diseaseid=int(phenotype_id), name=cols[DISEASE_NAME]
            )

            if cols[QUALIFIER] == "":
                disease.hpo.add(id_from_string(cols[HPO_ID]))
            elif cols[QUALIFIER] == "NOT":
                disease.negative_hpo.add(id_from_string(cols[HPO_ID]))


def _add_decipher_to_ontology(ontology: "pyhpo.OntologyClass") -> None:
    ontology._decipher_diseases = all_decipher_diseases()
    for decipher in ontology._decipher_diseases:
        for term_id in decipher.hpo:
            add_decipher_to_term(decipher, ontology[term_id])
        for term_id in decipher.negative_hpo:
            add_negative_decipher_to_term(decipher, ontology[term_id])


def _add_omim_to_ontology(ontology: "pyhpo.OntologyClass") -> None:
    ontology._omim_diseases = all_omim_diseases()
    for omim in ontology._omim_diseases:
        for term_id in omim.hpo:
            add_omim_to_term(omim, ontology[term_id])
        for term_id in omim.negative_hpo:
            add_negative_omim_to_term(omim, ontology[term_id])


def _add_orpha_to_ontology(ontology: "pyhpo.OntologyClass") -> None:
    ontology._orpha_diseases = all_orpha_diseases()
    for orpha in ontology._orpha_diseases:
        for term_id in orpha.hpo:
            add_orpha_to_term(orpha, ontology[term_id])
        for term_id in orpha.negative_hpo:
            add_negative_orpha_to_term(orpha, ontology[term_id])


def add_decipher_to_term(
    decipher: "pyhpo.annotations.DecipherDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add Decipher Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    decipher:
        Disease to add to term
    term:
        HPOTerm that is associated with diseease
    """
    if decipher in term.decipher_diseases:
        return None
    term.decipher_diseases.add(decipher)
    for parent in term.parents:
        add_decipher_to_term(decipher, parent)
    return None


def add_negative_decipher_to_term(
    decipher: "pyhpo.annotations.DecipherDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add excluded Decipher Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    decipher:
        Disease to exclude from term
    term:
        HPOTerm that is not associated with diseease
    """
    if decipher in term.decipher_excluded_diseases:
        return None
    term.decipher_excluded_diseases.add(decipher)
    for child in term.children:
        add_negative_decipher_to_term(decipher, child)
    return None


def add_omim_to_term(
    omim: "pyhpo.annotations.OmimDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add OMIM Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    omim:
        Disease to add to term
    term:
        HPOTerm that is associated with diseease
    """
    if omim in term.omim_diseases:
        return None
    term.omim_diseases.add(omim)
    for parent in term.parents:
        add_omim_to_term(omim, parent)
    return None


def add_negative_omim_to_term(
    omim: "pyhpo.annotations.OmimDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add excluded OMIM Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    omim:
        Disease to exclude from term
    term:
        HPOTerm that is not associated with diseease
    """
    if omim in term.omim_excluded_diseases:
        return None
    term.omim_excluded_diseases.add(omim)
    for child in term.children:
        add_negative_omim_to_term(omim, child)
    return None


def add_orpha_to_term(
    orpha: "pyhpo.annotations.OrphaDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add Orpha Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    orpha:
        Disease to add to term
    term:
        HPOTerm that is associated with diseease
    """
    if orpha in term.orpha_diseases:
        return None
    term.orpha_diseases.add(orpha)
    for parent in term.parents:
        add_orpha_to_term(orpha, parent)
    return None


def add_negative_orpha_to_term(
    orpha: "pyhpo.annotations.OrphaDisease", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add excluded Orpha Disease
    to an HPOTerm and all its parents

    Parameters
    ----------
    orpha:
        Disease to exclude from term
    term:
        HPOTerm that is not associated with diseease
    """
    if orpha in term.orpha_excluded_diseases:
        return None
    term.orpha_excluded_diseases.add(orpha)
    for child in term.children:
        add_negative_orpha_to_term(orpha, child)
    return None


def all_decipher_diseases() -> Set["DecipherDisease"]:
    return set(Decipher.keys())


def all_omim_diseases() -> Set["OmimDisease"]:
    return set(Omim.keys())


def all_orpha_diseases() -> Set["OrphaDisease"]:
    return set(Orpha.keys())



"""
Generic functions required for handling HPOTerms
They used to be part of HPOTerm, but refactored out to
have a leander HPOTerm class
"""

from typing import Iterator


def id_from_string(hpo_string: str) -> int:
    """
    Formats the HPO-type Term-ID into an integer id

    Parameters
    ----------
    hpo_string:
        HPO term ID.

        (e.g.: HP:000001)

    Returns
    -------
    int
        Integer representation of provided HPO ID

        (e.g.: 1)

    """
    idx = hpo_string.split("!")[0].strip()
    return int(idx.split(":")[1].strip())


def remove_outcommented_rows(fh: Iterator[str], ignorechar: str = "#") -> Iterator[str]:
    """
    Removes all rows from a filereader object that start
    with a comment character

    Parameters
    ----------
    fh:
        any object which supports the iterator protocol and
        returns a string each time its __next__() method is
        called — file objects and list objects are both suitable

    ignorechar:
        All lines starting with this character(s) will be ignored

    Yields
    ------
    row: str
        One row of the ``fh`` iterator
    """
    len_check = len(ignorechar)
    for row in fh:
        if row[0:len_check] != ignorechar:
            yield row



import csv
import os
from typing import Set

from pyhpo.annotations import Gene, GeneSingleton
from pyhpo.parser.generics import id_from_string, remove_outcommented_rows
import pyhpo


FILENAME = "phenotype_to_genes.txt"
# Cloumn mappings in phenotype_to_genes file
HPO_ID = 0
HGNC_ID = 2
GENE_SYMBOL = 3


def _parse_phenotype_to_gene_file(path: str) -> None:
    Gene.clear()
    filename = os.path.join(path, FILENAME)
    with open(filename) as fh:
        reader = csv.reader(
            remove_outcommented_rows(remove_outcommented_rows(fh), ignorechar="hpo_id"),
            delimiter="\t",
        )
        for cols in reader:
            gene = Gene(hgncid=int(cols[HGNC_ID]), symbol=cols[GENE_SYMBOL])
            gene.hpo.add(id_from_string(cols[HPO_ID]))


def _add_genes_to_ontology(ontology: "pyhpo.OntologyClass") -> None:
    ontology._genes = all_genes()
    for gene in ontology._genes:
        for term_id in gene.hpo:
            add_gene_to_term(gene, ontology[term_id])


def add_gene_to_term(
    gene: "pyhpo.annotations.GeneSingleton", term: "pyhpo.HPOTerm"
) -> None:
    """
    Recursive function to add Gene to an HPOTerm and all its parents

    Parameters
    ----------
    gene:
        Gene to add to term
    term:
        HPOTerm that is associated with the gene
    """
    if gene in term.genes:
        return None
    term.genes.add(gene)
    for parent in term.parents:
        add_gene_to_term(gene, parent)
    return None


def all_genes() -> Set["GeneSingleton"]:
    return set(Gene.keys())



from typing import Callable, Dict, List, Union, Tuple

try:
    from scipy.stats import hypergeom  # type: ignore[import]
except ImportError:
    print(
        "The pyhpo.stats module requires that you install scipy."
        + "\n\n#######################################################"
        + "\n\n#   ==> Please install scipy via `pip install scipy`  #"
        + "\n\n#######################################################\n\n"
    )
    raise ImportError()

import pyhpo
from pyhpo import Ontology
from pyhpo import HPOSet


def hypergeom_test(
    positive_samples: int, samples: int, positive_total: int, total: int
) -> float:
    """
    Wrapper function to call the scipy hypergeometric stats function

    Parameters
    ----------
        positive_samples: int
            Number of successes in the sample set (correctly drawn marbles)
        samples: int
            Total number of samples (number of drawn marbles)
        positive_total: int
            Number of positives in the reference set
            (number of positive marbles in the bag)
        total: int
            Total size of reference set
            (number of marbles in the bag)

    Returns
    -------
    float
        The hypergeometic enrichment score

    """
    return float(
        hypergeom.sf(
            positive_samples
            - 1,  # likelyhood of more than X, #see https://blog.alexlenail.me/understanding-and-implementing-the-hypergeometric-test-in-python-a7db688a7458  # noqa: E501
            total,
            positive_total,
            samples,
        )
    )


class HPOEnrichment:
    """
    Calculates the enrichment of HPO Terms in an Annotation set.

    You can use this class for the following example use cases:

    * You have a list of genes and want to see if some HPO terms are
      enriched in that group. (e.g. RNAseq differential gene expression)
    * You have a list of OMIM diseases and want to see if they have
      some underlying HPO symptom.

    Parameters
    ----------
    category: str
        String to declare if enrichment is done for genes or for OMIM diseases

        Options are:

        * **gene**
        * **omim**

    """

    def __init__(self, category: str) -> None:
        category_lookup = {"gene": Ontology.genes, "omim": Ontology.omim_diseases}
        self.hpos, self.total = self._hpo_count(
            category_lookup[category]  # type: ignore[arg-type]
        )

    def enrichment(
        self, method: str, annotation_sets: List["pyhpo.Annotation"]
    ) -> List[dict]:
        """
        Calculates the enrichment of HPO terms in the provided annotation set

        Parameters
        ----------
        method: str
            The statistical test for enrichment

            * **hypergeom** Hypergeometric distribution test

        annotation_sets: list of ``annoation``
            Every ``annotation`` item in the list must have an attribute
            ``hpos``, being a list of HPO-Term indicies

        Returns
        -------
        list of dict
            The enrichment of every HPO term in the ``annotation_sets`` list,
            sorted by descending enrichment. Every dict has the following keys:

            * **hpo**: :class:`.HPOTerm`
            * **count**: Number of appearances in the sets
            * **enrichment**: Enrichment score

        """
        list_counts, list_total = self._hpo_count(annotation_sets)
        res = [
            {
                "hpo": Ontology[hpo],
                "count": count,
                "enrichment": self._single_enrichment(method, hpo, count, list_total),
            }
            for hpo, count in list_counts.items()
        ]

        return sorted(res, key=lambda x: x["enrichment"])

    def _hpo_count(self, annotation_sets: List["pyhpo.Annotation"]) -> Tuple[dict, int]:
        """
        Counts the number of occurrenes of every HPO term
        in the ``annotation_sets``

        Parameters
        ----------
            annotation_set: list of ``annoation``
                Every ``annotation`` item in the list must have an attribute
                ``hpos``, returning an iterable of :class:`.HPOTerm`

        Returns
        -------
        tuple with following items:
            * Dict with
                * key: :class:`.HPOTerm`
                * value: int <Number of occurences>
            * Total number of HPO terms in set
        """
        hpos = {}
        for item in annotation_sets:
            for term in item.hpo:
                if term not in hpos:
                    hpos[term] = 0
                hpos[term] += 1
        return (hpos, sum(hpos.values()))

    def _single_enrichment(
        self,
        method: str,
        hpo_id: Union[int, "pyhpo.HPOTerm"],
        positives: int,
        samples: int,
    ) -> float:
        """
        Calculates the enrichment of a single HPO term compared to
        the reference set

        Parameters
        ----------
            method: str
                The statistical test for enrichment

                * **hypergeom** Hypergeometric distribution test

            hpo_id: int or :class:`.HPOTerm`
                ID of the HPO Term
            positives: int
                Number of successes in the sample set (correctly drawn marbles)
            samples: int
                Total number of samples (number of drawn marbles)

        Returns
        -------
        float
            The enrichment score
        """
        try:
            positive_total = self.hpos[hpo_id]
        except KeyError:
            raise RuntimeError(
                "The HPO term {} is not present in the "
                "reference population".format(hpo_id)
            )
        if method == "hypergeom":
            return hypergeom_test(positives, samples, positive_total, self.total)
        else:
            raise NotImplementedError("Enrichment method not implemented")


class EnrichmentModel:
    """
    Calculates the enrichment of annotations in an :class:`.HPOSet`.

    You can use this class for the following example use cases:

    * You have a set of HPOTerms and want to find the
      most likely causative gene
    * You have a set of HPOTerms and want to find the underlying disease

    Parameters
    ----------
    category: str
        String to declare if enrichment is done for genes or for OMIM diseases

        Options are:

        * **gene**
        * **omim**
        * **orpha**
        * **decipher**

    """

    attribute_lookup: Dict[str, Callable] = {
        "gene": lambda x: x.genes,
        "omim": lambda x: x.omim_diseases,
        "orpha": lambda x: x.orpha_diseases,
        "decipher": lambda x: x.decipher_diseases,
    }

    def __init__(self, category: str) -> None:
        self.attribute = self.attribute_lookup[category]
        self.base_count, self.total = self._population_count(HPOSet(Ontology))

    def enrichment(self, method: str, hposet: HPOSet) -> List[dict]:
        """
        Calculates the enrichment of annotations in the provided HPOSet

        Parameters
        ----------
        method: str
            The statistical test for enrichment

            * **hypergeom** Hypergeometric distribution test

        hposet: :class:`.HPOSet`

        Returns
        -------
        list of dict
            The enrichment of every annotation item sorted by
            descending enrichment. Every dict has the following keys:

            * **item**: Gene or OMIM or Decipher annotation item
            * **count**: Number of appearances in the sets
            * **enrichment**: Enrichment score

        """
        list_counts, list_total = self._population_count(hposet)
        res = [
            {
                "item": item,
                "count": count,
                "enrichment": self._single_enrichment(method, item, count, list_total),
            }
            for item, count in list_counts.items()
        ]
        return sorted(res, key=lambda x: x["enrichment"])

    def _population_count(self, hopset: HPOSet) -> Tuple[dict, int]:
        """
        Counts the number of occurrenes of every annotation item
        in the HPOSet

        Parameters
        ----------
            hposet: :class:`.HPOSet`

        Returns
        -------
        tuple with following items:
            * Dict with
                * key: Annotation Item
                * value: int <Number of occurences>
            * Total number of annotations in set
        """
        population = {}
        for term in hopset:
            for item in self.attribute(term):
                if item not in population:
                    population[item] = 0
                population[item] += 1
        return population, sum(population.values())

    def _single_enrichment(
        self, method: str, item_id: int, positives: int, samples: int
    ) -> float:
        """
        Calculates the enrichment of annotations in an HPO set

        Parameters
        ----------
            method: str
                The statistical test for enrichment

                * **hypergeom** Hypergeometric distribution test

            item_id: int
                ID of the Annotation
            positives: int
                Number of successes in the sample set (correctly drawn marbles)
            samples: int
                Total number of samples (number of drawn marbles)

        Returns
        -------
        float
            The enrichment score
        """
        try:
            positive_total = self.base_count[item_id]
        except KeyError:
            raise RuntimeError(
                "The item {} is not present in the "
                "reference population".format(item_id)
            )
        if method == "hypergeom":
            return hypergeom_test(positives, samples, positive_total, self.total)
        else:
            raise NotImplementedError("Enrichment method not implemented")



import math
from typing import List

from pyhpo.similarity.base import SimilarityBase
import pyhpo


class Resnik(SimilarityBase):
    """
    Based on Resnik P, Proceedings of the 14th IJCAI, (1995)
    """

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        sim = 0.0
        for term in term1.common_ancestors(term2):
            ic = term.information_content[kind]
            if ic > sim:
                sim = ic

        return sim


class Lin(SimilarityBase):
    """
    Based on Lin D, Proceedings of the 15th ICML, (1998)
    """

    dependencies: List[str] = ["resnik"]

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        ic_t1 = term1.information_content[kind]
        ic_t2 = term2.information_content[kind]
        try:
            return (2 * dependencies[0]) / (ic_t1 + ic_t2)
        except ZeroDivisionError:
            return 0.0


class JC(SimilarityBase):
    """
    This method is the same as the source code in
    the R package ``hposim``

    .. code-block:: r

        res= - 1/ ( 1 + 2*IC[IC[,1]==an,3] - IC[IC[,1]==term1,3]
        - IC[IC[,1]==term2,3] )

    .. note::

        This method is *not recommended*.

        See :func:`pyhpo.term._jc_similarity_score_2`
        for an alternative way to calculate Jiang & Conrath
    """

    dependencies: List[str] = ["resnik"]

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        if term1 == term2:
            return 1.0

        ic_t1 = term1.information_content[kind]
        ic_t2 = term2.information_content[kind]

        return -1 / (1 + (2 * dependencies[0]) - ic_t1 - ic_t2)


class JC2(SimilarityBase):
    """
    Jiang & Conrath similarity Score, based on
    Jiang J, Conrath D, ROCLING X, (1997) and
    Deng Y, et. al., PLoS One, (2015)

    This method is the same as the description
    in the paper for the R package ``hposim``
    Deng Y, et. al., PLoS One, (2015)

    ::

        sim[JC](t1,t2) = 1-(IC(t1)+IC(t2)−2×IC(t[MICA]))

    """

    dependencies: List[str] = ["resnik"]

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        if term1 == term2:
            return 1.0

        ic_t1 = term1.information_content[kind]
        ic_t2 = term2.information_content[kind]

        return 1 - (ic_t1 + ic_t2 - (2 * dependencies[0]))


class Relevance(SimilarityBase):
    """
    Based on Schlicker A, et.al., BMC Bioinformatics, (2006)
    """

    dependencies: List[str] = ["resnik", "lin"]

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        return dependencies[1] * (1 - (math.exp(dependencies[0] * -1)))


class InformationCoefficient(SimilarityBase):
    """
    Based on Li B, et. al., arXiv, (2010)
    """

    dependencies: List[str] = ["resnik", "lin"]

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        return dependencies[1] * (1 - (1 / (1 + dependencies[0])))


class GraphIC(SimilarityBase):
    """
    Graph based Information coefficient, based on
    Deng Y, et. al., PLoS One, (2015)
    """

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        if term1 == term2:
            return 1.0

        common = sum(
            [x.information_content[kind] for x in term1.common_ancestors(term2)]
        )
        union = sum(
            [
                x.information_content[kind]
                for x in (term1.all_parents | term2.all_parents)
            ]
        )

        try:
            return common / union
        except ZeroDivisionError:
            return 0.0


class Distance(SimilarityBase):
    """
    actual distance (number of hpos) between Terms
    """

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        try:
            dist = term1.path_to_other(term2)[0]
        except IndexError:
            # No path found -> This can only happen if one of the
            # terms is obsolete or otherwise not part of the Ontology
            return 0.0

        return 1 / (dist + 1)


def register_defaults(simscore: "pyhpo.similarity.base._Similarity") -> None:
    simscore.register("resnik", Resnik)
    simscore.register("lin", Lin)
    simscore.register("jc", JC)
    simscore.register("jc2", JC2)
    simscore.register("rel", Relevance)
    simscore.register("ic", InformationCoefficient)
    simscore.register("graphic", GraphIC)
    simscore.register("dist", Distance)



# -*- coding: utf-8 -*-

"""PyHPO: A library for handling the Human Phenotype Ontology"""

from pyhpo.term import HPOTerm
from pyhpo.ontology import Ontology, OntologyClass
from pyhpo.set import HPOSet
from pyhpo.annotations import Annotation, GeneSingleton, DiseaseSingleton
from pyhpo.annotations import OmimDisease, OrphaDisease, DecipherDisease
from pyhpo import config

# The following info will be used by setup.py and sphinx documentation
__author__ = "Jonas Marcello"
__version__ = "3.2.6"

__all__ = (
    "Annotation",
    "HPOTerm",
    "HPOSet",
    "Ontology",
    "OntologyClass",
    "GeneSingleton",
    "DiseaseSingleton",
    "OmimDisease",
    "OrphaDisease",
    "DecipherDisease",
    "config",
)



import warnings
from typing import Iterable, Set, List, Iterator, Union, Tuple

import pyhpo
from pyhpo.ontology import Ontology
from pyhpo.matrix import Matrix


class HPOSet(set):
    def __init__(self, items: Iterable["pyhpo.HPOTerm"]) -> None:
        set.__init__(self, items)
        self._list: List["pyhpo.HPOTerm"] = list(items)

    def add(self, item: "pyhpo.HPOTerm") -> None:
        """
        Overwrites ``set.add`` to ensure we keep the
        ``self._list`` property updated as well.
        """
        if item not in self:
            set.add(self, item)
            self._list.append(item)

    def update(  # type: ignore[override]
        self, items: Iterable["pyhpo.HPOTerm"]
    ) -> None:
        """
        Overwrites ``set.update`` to ensure we keep the
        ``self._list`` property updated as well.
        """
        for item in items:
            self.add(item)

    def child_nodes(self) -> "HPOSet":
        """
        Return a new HPOSet tha contains only
        the most specific HPO term for each subtree

        It basically will return only HPO terms
        that do not have descendant HPO terms
        present in the set

        Returns
        -------
        HPOSet
            HPOSet instance that contains only the most specific
            child nodes of the current HPOSet
        """
        counter = {term.id: 0 for term in self}
        for child, parent in self.combinations():
            if child.child_of(parent):
                counter[parent.id] += 1
        return HPOSet([term for term in self if counter[term.id] == 0])

    def remove_modifier(self) -> "HPOSet":
        """
        Removes all modifier terms. By default, this includes

        * ``Mode of inheritance: 'HP:0000005'``
        * ``Clinical modifier: 'HP:0012823'``
        * ``Frequency: 'HP:0040279'``
        * ``Clinical course: 'HP:0031797'``
        * ``Blood group: 'HP:0032223'``
        * ``Past medical history: 'HP:0032443'``


        Returns
        -------
        HPOSet
            HPOSet instance that contains only
            ``Phenotypic abnormality`` HPO terms

        """

        return HPOSet([term for term in self if not term.is_modifier])

    def replace_obsolete(self, verbose: bool = False) -> "HPOSet":
        """
        Replaces obsolete terms with the replacement term

        .. warning::

            Not all obsolete terms have a replacement. Obsolete terms
            without replacements will be removed from the set.

        Parameters
        ----------
        verbose: bool, default: ``False``
            Print warnings if an obsolete term does not have a replacement.

        Returns
        -------
        HPOSet
            A new HPOSet

        """
        ids: Set["pyhpo.HPOTerm"] = set()
        for term in self:
            if term.is_obsolete:
                try:
                    replaced = Ontology.get_hpo_object(term.replaced_by)
                    ids.add(replaced)
                except TypeError:
                    warnings.warn(
                        "The term {} is obsolete and has no replacement.".format(term),
                        UserWarning,
                    )

            else:
                ids.add(term)
        return HPOSet(ids)

    def all_genes(self) -> Set["pyhpo.GeneSingleton"]:
        """
        Calculates the union of the genes
        attached to the HPO Terms in this set

        Returns
        -------
        set of :class:`annotations.Gene`
            Set of all genes associated with the HPOTerms in the set
        """
        genes = set()
        for term in self:
            genes.update(term.genes)
        return genes

    def omim_diseases(self) -> Set["pyhpo.OmimDisease"]:
        """
        Calculates the union of the Omim diseases
        attached to the HPO Terms in this set

        Returns
        -------
        set of :class:`annotations.Omim`
            Set of all Omim diseases associated with the HPOTerms in the set
        """
        omims = set()
        for term in self:
            omims.update(term.omim_diseases)
        return omims

    def orpha_diseases(self) -> Set["pyhpo.OrphaDisease"]:
        """
        Calculates the union of the Omim diseases
        attached to the HPO Terms in this set

        Returns
        -------
        set of :class:`annotations.Omim`
            Set of all Omim diseases associated with the HPOTerms in the set
        """
        orphas = set()
        for term in self:
            orphas.update(term.orpha_diseases)
        return orphas

    def decipher_diseases(self) -> Set["pyhpo.DecipherDisease"]:
        """
        Calculates the union of the Omim diseases
        attached to the HPO Terms in this set

        Returns
        -------
        set of :class:`annotations.Omim`
            Set of all Omim diseases associated with the HPOTerms in the set
        """
        deciphers = set()
        for term in self:
            deciphers.update(term.decipher_diseases)
        return deciphers

    def information_content(self, kind: str = "") -> dict:
        """
        Gives back basic information content stats about the
        HPOTerms within the set

        Parameters
        ----------
        kind: str, default: ``omim``
            Which kind of information content should be calculated.
            Options are ['omim', 'orpha', 'decipher', 'gene']


        Returns
        -------
        dict
            Dict with the following items

            * **mean** - float - Mean information content
            * **max** - float - Maximum information content value
            * **total** - float - Sum of all information content values
            * **all** - list of float -
              List with all information content values
        """
        if not kind:
            kind = "omim"

        res = {
            "mean": None,
            "total": 0,
            "max": 0,
            "all": [term.information_content[kind] for term in self],
        }
        res["total"] = sum(res["all"])  # type: ignore
        res["max"] = max(res["all"])  # type: ignore
        res["mean"] = res["total"] / len(self)  # type: ignore
        return res

    def variance(self) -> Tuple[float, int, int, List[int]]:
        """
        Calculates the distances between all its term-pairs. It also provides
        basic calculations for variances among the pairs.

        Returns
        -------
        tuple of (int, int, int, list of int)
            Tuple with the variance metrices

            * **float** Average distance between pairs
            * **int** Smallest distance between pairs
            * **int** Largest distance between pairs
            * **list of int** List of all distances between pairs
        """

        distances = []
        for term_a, term_b in self.combinations_one_way():
            distances.append(term_a.path_to_other(term_b)[0])

        if len(distances):
            return (
                sum(distances) / len(distances),
                min(distances),
                max(distances),
                distances,
            )
        else:
            return (0, 0, 0, [])

    def combinations(self) -> Iterator[Tuple["pyhpo.HPOTerm", "pyhpo.HPOTerm"]]:
        """
        Helper generator function that returns all possible two-pair
        combination between all its terms

        This function is direction dependent. That means that every
        pair will appear twice. Once for each direction

        .. seealso:: :func:`pyhpo.set.HPOSet.combinations_one_way`

        Yields
        ------
        Tuple of :class:`term.HPOTerm`
            Tuple containing the follow items

            * **HPOTerm** instance 1 of the pair
            * **HPOTerm** instance 2 of the pair

        Examples
        --------
            ::

                ci = HPOSet([term1, term2, term3])
                ci.combinations()

                # Output:
                [
                    (term1, term2),
                    (term1, term3),
                    (term2, term1),
                    (term2, term3),
                    (term3, term1),
                    (term3, term2)
                ]

        """
        for term_a in self._list:
            for term_b in self._list:
                if term_a == term_b:
                    continue
                yield (term_a, term_b)

    def combinations_one_way(self) -> Iterator[Tuple["pyhpo.HPOTerm", "pyhpo.HPOTerm"]]:
        """
        Helper generator function that returns all possible two-pair
        combination between all its terms

        This methow will report each pair only once

        .. seealso:: :func:`pyhpo.set.HPOSet.combinations`

        Yields
        ------
        Tuple of :class:`term.HPOTerm`
            Tuple containing the follow items

            * **HPOTerm** instance 1 of the pair
            * **HPOTerm** instance 2 of the pair

        Example
        -------
            ::

                ci = HPOSet([term1, term2, term3])
                ci.combinations()

                # Output:
                [
                    (term1, term2),
                    (term1, term3),
                    (term2, term3)
                ]

        """
        for i, term_a in enumerate(self._list):
            for term_b in self._list[i + 1 :]:
                yield (term_a, term_b)

    def similarity(
        self,
        other: "HPOSet",
        kind: str = "",
        method: str = "",
        combine: str = "funSimAvg",
    ) -> float:
        """
        Calculates the similarity to another HPOSet
        According to Robinson et al, American Journal of Human Genetics, (2008)
        and Pesquita et al, BMC Bioinformatics, (2008)

        Parameters
        ----------
        other: HPOSet
            Another HPOSet to measure the similarity to

        kind: str, default ``''``
            Which kind of information content should be calculated.
            Options are ['omim', 'orpha', 'decipher', 'gene']
            See :func:`pyhpo.HPOTerm.similarity_score` for options

        method: string, default ``''``
            The method to use to calculate the similarity.
            See :func:`pyhpo.HPOTerm.similarity_score` for options

            Additional options:

            * **equal** - Calculates exact matches between both sets

        combine: string, default ``funSimAvg``
            The method to combine similarity measures.

            Available options:

            * **funSimAvg** - Schlicker A, BMC Bioinformatics, (2006)
            * **funSimMax** - Schlicker A, BMC Bioinformatics, (2006)
            * **BMA** - Deng Y, et. al., PLoS One, (2015)

        Returns
        -------
        float
            The similarity score to the other HPOSet


        Raises
        ------
        RuntimeError
            The specified ``method`` or ``combine`` does not exist
        NotImplementedError
            This error can only occur with custom Similarity-Score
            methods that do not have a ``similarity`` method defined.
        AttributeError
            The information content for ``kind`` does not exist
        """
        if method == "equal":
            return self._equality_score(other)

        score_matrix = HPOSet._sim_score(self, other, kind, method)

        row_maxes = [max([v for v in row]) for row in score_matrix.rows]

        col_maxes = [max([v for v in col]) for col in score_matrix.columns]

        try:
            if combine == "funSimAvg":
                return (
                    sum(row_maxes) / len(row_maxes) + sum(col_maxes) / len(col_maxes)
                ) / 2

            if combine == "funSimMax":
                return max(
                    [sum(row_maxes) / len(row_maxes), sum(col_maxes) / len(col_maxes)]
                )

            if combine == "BMA":
                return (sum(row_maxes) + sum(col_maxes)) / (
                    len(row_maxes) + len(col_maxes)
                )
        except ZeroDivisionError:
            return 0

        raise RuntimeError("Invalid combine method specified")

    def _equality_score(self, other: "HPOSet") -> float:
        """
        Returns an equality similarity score.
        Only exact matches between both sets are counted
        and the fraction of exact matches is returned.
        A score of 1 means both sets match 100%,
        0.5 means only half the terms have an exact match.

        This method does not take advantage of the ontology
        and does not take distance measures into account.

        Parameters
        ----------
        other: HPOSet
            Another HPOSet to measure the similarity to

        Returns
        -------
        float
            The similarity score to the other HPOSet

        """
        if not len(self) or not len(other):
            return 0

        matches = 0
        for term1 in self:
            for term2 in other:
                if term1 == term2:
                    matches += 1
        return matches / max([len(self), len(other)])

    @staticmethod
    def _sim_score(
        set1: "HPOSet", set2: "HPOSet", kind: str = "", method: str = ""
    ) -> Matrix:
        """
        Calculates similarity matrix between HPOSets

        .. warning::

           This method should not be used by itself.
           Use :func:`pyhpo.set.HPOSet.similarity` instead.

        Parameters
        ----------
        set1: HPOSet
            One HPOSet to measure the similarity from
        set2: HPOSet
            Another HPOSet to measure the similarity to

        kind: str
            Which kind of information content should be calculated.
            See :function:`pyhpo.HPOTerm.similarity_score` for options

        method: string
            The method to use to calculate the similarity.
            See :function:`pyhpo.HPOTerm.similarity_score` for options

        Returns
        -------
        float
            The one-way similarity from one to the other HPOSet

        """

        if not len(set1) or not len(set2):
            return Matrix(0, 0)

        scores = []
        for set1_term in set1:
            for set2_term in set2:
                scores.append(set1_term.similarity_score(set2_term, kind, method))

        return Matrix(len(set1), len(set2), scores)

    @classmethod
    def from_queries(cls, queries: Iterable[Union[int, str]]) -> "HPOSet":
        """
        Builds an HPO set by specifying a list of queries to run on the
        :class:`pyhpo.ontology.Ontology`

        Parameters
        ----------
        queries: list of (string or int)
            The queries to be run the identify the HPOTerm from the ontology

        Returns
        -------
        :class:`pyhpo.set.HPOSet`
            A new HPOset

        Examples
        --------
            ::

                ci = HPOSet([
                    'Scoliosis',
                    'HP:0001234',
                    12
                ])

        """
        return cls([Ontology.get_hpo_object(query) for query in queries])

    @classmethod
    def from_serialized(cls, pickle: str) -> "HPOSet":
        """
        Re-Builds an HPO set from a serialized HPOSet object

        Parameters
        ----------
        pickle: str
            The serialized HPOSet object

        Returns
        -------
        :class:`pyhpo.set.HPOSet`
            A new HPOset

        Examples
        --------
            ::

                ci = HPOSet(ontology, '12+24+66628')

        """
        return cls([Ontology.get_hpo_object(int(query)) for query in pickle.split("+")])

    def serialize(self) -> str:
        """
        Creates a string serialization that can be used to
        rebuild the same HPOSet via :func:`pyhpo.set.HPOSet.from_serialized`

        Returns
        -------
        str
            A string representation of the HPOSet

        """
        ids = [str(x) for x in sorted([int(x) for x in self])]
        return "+".join(ids)

    def toJSON(self, verbose: bool = False) -> List[dict]:
        """
        Creates a JSON-like object of the HPOSet

        Parameters
        ----------
        verbose: bool, default ``False``
            Include extra properties of the HPOTerm

        Returns
        -------
        list of dict
            a list of HPOTerm dict objects

        """
        return [t.toJSON(verbose) for t in self._list]

    def __str__(self) -> str:
        return "{}: {}".format(
            self.__class__.__name__, ", ".join([x.name for x in self._list])
        )

    def __repr__(self) -> str:
        return '{}.from_serialized("{}")'.format(
            self.__class__.__name__, self.serialize()
        )


class BasicHPOSet(HPOSet):
    """
    Child of :class:`.HPOSet` that automatically:

    * removes parent terms
    * removes modifier terms
    * replaces obsolete terms
    """

    def __init__(self, items: Iterable["pyhpo.HPOTerm"]) -> None:
        HPOSet.__init__(self, [])
        for item in items:
            self.add(item)

    def add(self, item: "pyhpo.HPOTerm") -> None:
        """
        Overwrites ``set.add`` to ensure we keep the
        ``self._list`` property updated and
        don't add modifiers, obsolete or parent terms
        as well
        """
        if item in self:
            return None
        if item.is_modifier:
            return None
        for term in self:
            if item.parent_of(term):
                return None
        for p in item.all_parents:
            if p in self:
                self.remove(p)
        set.add(self, item)
        self._list.append(item)



import pyhpo
from pyhpo.parser import diseases
from pyhpo.parser import genes


def build_ontology_annotations(
    data_folder: str, ontology: "pyhpo.OntologyClass"
) -> None:
    """
    Builds all annotations and associated genes and diseases
    to the HPO terms in the Ontology.

    Parameters
    ----------
    data_folder:
        The full path to the folder that contains the master data files

    ontology:
        The ontology of HPO terms
    """
    genes._parse_phenotype_to_gene_file(data_folder)
    genes._add_genes_to_ontology(ontology)

    diseases._parse_phenotype_hpoa_file(data_folder)
    diseases._add_decipher_to_ontology(ontology)
    diseases._add_omim_to_ontology(ontology)
    diseases._add_orpha_to_ontology(ontology)



import os
import math
import warnings
from typing import List, Set, Tuple, Optional, Union, Dict, Iterator

try:
    import pandas as pd  # type: ignore
except ImportError:
    warnings.warn(
        "Some functionality requires pandas, which is currently not available",
        UserWarning,
    )

import pyhpo
from pyhpo import HPOTerm
from pyhpo.parser import build_ontology_annotations
from pyhpo.parser.obo import terms_from_file
from pyhpo.parser.generics import id_from_string


class OntologyClass:
    """
    A linked and indexed list of interconnected :class:`HPOTerm` s.

    Attributes
    ----------
    genes: set
        Set of all genes associated with the HPOTerms
    omim_diseases: set
        Set of all OMIM-diseases associated with the HPOTerms
    omim_excluded_diseases: set
        Set of all excluded OMIM-diseases associated with the HPOTerms

    """

    def __call__(
        self, data_folder: Optional[str] = None, from_obo_file: bool = True
    ) -> "OntologyClass":
        self.metadata: List[str] = []
        self._map: Dict[int, HPOTerm] = {}
        self._genes: Set["pyhpo.GeneSingleton"] = set()
        self._omim_diseases: Set["pyhpo.OmimDisease"] = set()
        self._orpha_diseases: Set["pyhpo.OrphaDisease"] = set()
        self._decipher_diseases: Set["pyhpo.DecipherDisease"] = set()

        if data_folder is None:
            data_folder = os.path.join(os.path.dirname(__file__), "data")

        if from_obo_file:
            self._load_from_obo_file(data_folder)

        return self

    def get_hpo_object(self, query: Union[int, str]) -> HPOTerm:
        """
        Matches a single HPO term based on its name, synonym or id

        Parameters
        ----------
        query: str or int

            * **str** HPO term ``Scoliosis``
            * **str** synonym ``Curved spine``
            * **str** HPO-ID ``HP:0002650``
            * **int** HPO term id ``2650``

        Returns
        -------
        HPOTerm
            A single matching HPO term instance

        Raises
        ------
        RuntimeError
            No HPO term is found for the provided query
        TypeError
            The provided query is an unsupported type and can't be properly
            converted
        ValueError
            The provided HPO ID cannot be converted to the correct
            integer representation

        Example
        -------
            ::

                # Search by ID (int)
                >>> ontology.get_hpo_object(3)
                HP:0000003 | Multicystic kidney dysplasia

                # Search by HPO-ID (string)
                >>> ontology.get_hpo_object('HP:0000003')
                HP:0000003 | Multicystic kidney dysplasia

                # Search by term (string)
                >>> ontology.get_hpo_object('Multicystic kidney dysplasia')
                HP:0000003 | Multicystic kidney dysplasia

                # Search by synonym (string)
                >>> ontology.get_hpo_object('Multicystic renal dysplasia')
                HP:0000003 | Multicystic kidney dysplasia

        """
        res: Optional[HPOTerm] = None
        if isinstance(query, str):
            if query.startswith("HP:"):
                try:
                    res = self[id_from_string(query)]
                except ValueError as err:
                    raise ValueError(f"Invalid id: {query}") from err
                except KeyError:
                    pass
            else:
                try:
                    res = self.synonym_match(query)
                except RuntimeError:
                    pass

        elif isinstance(query, int):
            try:
                res = self[query]
            except KeyError:
                pass

        else:
            raise TypeError('Invalid type {} for parameter "query"'.format(type(query)))

        if res:
            return res
        else:
            raise RuntimeError("Unknown HPO term")

    def match(self, query: str) -> HPOTerm:
        """
        Matches a single HPO term based on its name

        Parameters
        ----------
        query: str
            HPO term to match
            e.g: Abnormality of the nervous system

        Returns
        -------
        HPOTerm
            A single matching HPO term instance
        """

        for term in self:
            if query == term.name:
                return term
        raise RuntimeError("No HPO entry with name {}".format(query))

    def path(
        self, query1: Union[int, str], query2: Union[int, str]
    ) -> Tuple[int, Tuple[HPOTerm, ...], int, int]:
        """
        Returns the shortest connection between
        two HPO terms

        Parameters
        ----------
        query1: str or int
            HPO term 1, synonym or HPO-ID (HP:00001) to match
            HPO term id (Integer based)
            e.g: Abnormality of the nervous system
        query2: str or int
            HPO term 2, synonym or HPO-ID (HP:00001) to match
            HPO term id (Integer based)
            e.g: Abnormality of the nervous system

        Returns
        -------
        int
            Length of path
        tuple
            Tuple of HPOTerms in the path
        int
            Number of steps from term-1 to the common parent
        int
            Number of steps from term-2 to the common parent

        """

        term1 = self.get_hpo_object(query1)
        term2 = self.get_hpo_object(query2)
        return term1.path_to_other(term2)

    def search(self, query: str) -> Iterator[HPOTerm]:
        """
        Iterator function for substring search
        for terms and synonyms in the ontology

        Parameters
        ----------
        query:  str
            Term to search for

        Yields
        ------
        HPOTerm
            Every matching HPO term instance
        """
        for term in self:
            if (query.lower() in term.name.lower()) or (
                self.synonym_search(term, query)
            ):
                yield term

    def synonym_match(self, query: str) -> HPOTerm:
        """
        Searches for actual and synonym term matches
        If a match is found in any term, that one is returned
        If no actual match is found, the first match
        with synonyms is considered

        Parameters
        ----------
        query:  str
            Term to search for

        Returns
        -------
        HPOTerm
            A single HPO term instance

        """
        synonym_hit = None
        for term in self:
            if query == term.name:
                return term
            if not synonym_hit and query in term.synonym:
                synonym_hit = term

        if synonym_hit:
            return synonym_hit
        raise RuntimeError("No HPO entry with term or synonym {}".format(query))

    def synonym_search(self, term: HPOTerm, query: str) -> bool:
        """
        Indicates whether a term has a synonym that contains the query.
        Substring search in synonym

        Parameters
        ----------
        term:  HPOTerm
            Term to check
        query:  str
            Synonym substring

        Returns
        -------
        bool
            Indicates if ``term`` contains ``query`` as a synonym
        """

        for synonym in term.synonym:
            if query.lower() in synonym.lower():
                return True
        return False

    def to_dataframe(self) -> "pd.DataFrame":
        """
        Creates a Pandas DataFrame from the most important features

        Each HPO term is one row, the features are present in columns

        Returns
        -------
        :class:`DataFrame`
            The DataFrame of HPO-Terms and their
            attributes in the following columns

            * **id** ``str`` The HPO Term ID "HP:0000003" (used as index)
            * **name** ``str`` The HPO Term name "Multicystic kidney dysplasia"
            * **parents** ``str`` Concatenated list of direct parents of
              HPO terms. Separated by ``|``
            * **children** ``str`` Concatenated list of direct children of
              HPO terms. Separated by ``|``
            * **ic_omim** ``float`` Information-content
              (based on associated OMIM diseases)
            * **ic_gene** ``float`` Information-content
              (based on associated genes)
            * **dTop_l** ``int`` Maximum distance to root term
              (via :func:`pyhpo.term.longest_path_to_root`)
            * **dTop_s** ``int`` Shortest distance to root term
              (via :func:`pyhpo.term.shortest_path_to_root`)
            * **dBottom** ``int`` Longest graph of children nodes
              (via :func:`pyhpo.term.longest_path_to_bottom`)
            * **genes** ``str`` Concatenated list of associated
              genes. Separated by ``|``
            * **diseases** ``str`` Concatenated list of associated
              OMIM diseases. Separated by ``|``
        """

        data: Dict[str, List[Union[float, int, str]]] = {
            "id": [],
            "name": [],
            "parents": [],
            "children": [],
            "ic_omim": [],
            "ic_orpha": [],
            "ic_decipher": [],
            "ic_gene": [],
            "dTop_l": [],
            "dTop_s": [],
            "dBottom": [],
            "genes": [],
            "omim": [],
            "orpha": [],
            "decipher": [],
        }

        # This is not the most elegant way to generate a DataFrame
        # But it works
        for term in self:
            data["id"].append(term.id)
            data["name"].append(term.name)
            data["parents"].append("|".join([x.id for x in term.parents]))
            data["children"].append("|".join([x.id for x in term.children]))
            data["ic_omim"].append(term.information_content.omim)
            data["ic_orpha"].append(term.information_content.orpha)
            data["ic_decipher"].append(term.information_content.decipher)
            data["ic_gene"].append(term.information_content.gene)
            data["dTop_l"].append(term.longest_path_to_root())
            data["dTop_s"].append(term.shortest_path_to_root())
            data["dBottom"].append(term.longest_path_to_bottom())
            data["genes"].append("|".join([str(x) for x in term.genes]))
            data["omim"].append("|".join([str(x) for x in term.omim_diseases]))
            data["orpha"].append("|".join([str(x) for x in term.omim_diseases]))
            data["decipher"].append("|".join([str(x) for x in term.omim_diseases]))

        return pd.DataFrame(data).set_index("id")

    @property
    def genes(self) -> Set["pyhpo.GeneSingleton"]:
        return self._genes

    @property
    def decipher_diseases(self) -> Set["pyhpo.DecipherDisease"]:
        return self._decipher_diseases

    @property
    def omim_diseases(self) -> Set["pyhpo.OmimDisease"]:
        return self._omim_diseases

    @property
    def orpha_diseases(self) -> Set["pyhpo.OrphaDisease"]:
        return self._orpha_diseases

    def _load_from_obo_file(self, data_folder: str) -> None:
        """
        Reads an obo file line by line to add
        HPO terms to the Ontology

        Attributes
        ----------
        data_folder: str
            Full path to folder where master data is stored

        """
        for term in terms_from_file(data_folder):
            self._append(HPOTerm(**term))

        self._connect_all()
        build_ontology_annotations(data_folder, self)  # type: ignore
        self._add_information_content()

    def _append(self, item: HPOTerm) -> None:
        """
        Adds one HPO term to the ontology
        """
        self._map[item.index] = item

    def _connect_all(self) -> None:
        """
        Connects all parent-child associations in the Ontology
        Called by default after loading the ontology from a file
        """
        for term in self._map.values():
            for parent_id in term.parent_ids():
                parent = self[parent_id]
                term.parents.add(parent)
                parent.children.add(term)

        # Build caches of hierarchy to speed up performance
        for term in self._map.values():
            term.all_parents

    def _add_information_content(self) -> None:
        """
        Calculates the information content for each HPO Term
        According to Robinson et al, American Journal of Human Genetics, 2008
        https://www.sciencedirect.com/science/article/pii/S0002929708005351

        Returns
        -------
        None
            None
        """
        total_omim_diseases = len(self.omim_diseases)
        total_orpha_diseases = len(self.orpha_diseases)
        total_decipher_diseases = len(self.decipher_diseases)
        total_genes = len(self.genes)
        for term in self:
            p_omim = len(term.omim_diseases) / total_omim_diseases
            p_orpha = len(term.orpha_diseases) / total_orpha_diseases
            p_decipher = len(term.decipher_diseases) / total_decipher_diseases
            p_gene = len(term.genes) / total_genes
            if p_omim == 0:
                term.information_content.omim = 0
            else:
                term.information_content.omim = -math.log(p_omim)

            if p_orpha == 0:
                term.information_content.orpha = 0
            else:
                term.information_content.orpha = -math.log(p_orpha)

            if p_decipher == 0:
                term.information_content.decipher = 0
            else:
                term.information_content.decipher = -math.log(p_decipher)

            if p_gene == 0:
                term.information_content.gene = 0
            else:
                term.information_content.gene = -math.log(p_gene)

    def __getitem__(self, key: int) -> HPOTerm:
        try:
            return self._map[key]
        except KeyError as e:
            raise KeyError("No HPOTerm for index {}".format(key)) from e

    def __iter__(self) -> Iterator[HPOTerm]:
        return iter(self._map.values())

    def __len__(self) -> int:
        return len(self._map.keys())


Ontology: OntologyClass = OntologyClass()



from typing import Any, ClassVar, Dict, Set, Union

from pydantic import BaseModel


class Annotation(BaseModel):
    id: int
    name: str
    hpo: Set[int] = set()
    _hash: int
    _json_keys = set(["id", "name"])

    def __init__(self, **kwargs: Union[int, str]) -> None:
        super().__init__(**kwargs)
        self._hash = hash((self.id, self.name))

    def toJSON(self, verbose: bool = False) -> dict:
        """
        Backwards compatibility method
        BaseModel include ``.json`` method

        Parameters
        ----------
        verbose: bool, default: ``False``
            Return all associated HPOTerms

        Returns
        -------
        dict
            A dict with the following keys
            (additional keys might be present, depending on the class)

            * **id** - The HGNC ID
            * **name** - The gene symbol
            * **hpo** - (If ``verbose == True``):
              set of :class:`pyhpo.term.HPOTerm`
        """
        res = {}
        for key in self._json_keys:
            res[key] = self.__getattribute__(key)
        if verbose:
            res["hpo"] = self.hpo
        return res

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, int):
            return self.id == other

        if isinstance(other, str):
            return self.name == other

        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return self._hash

    def __str__(self) -> str:
        return self.name


class GeneSingleton(Annotation):
    """
    An instance of ``GeneSingleton`` represents a single gene.

    .. note::

        ``GeneSingleton`` should never be initiated directly,
        but only via :class:`.GeneDict`
        to ensure that every gene is only created once.

    Attributes
    ----------
    id: int
        HGNC gene ID

    name: str
        HGNC gene synbol

    symbol: str
        HGNC gene symbol (alias of :attr:`.GeneSingleton.name`)

    hpo: set of :class:`pyhpo.term.HPOTerm`
        all HPOTerms associated to the gene

    Parameters
    ----------
    id: int
        HGNC gene ID
    name: str
        HGNC gene synbol
    """

    _json_keys = set(["id", "name", "symbol"])

    @property
    def symbol(self) -> str:
        return self.name


class GeneDict(dict):
    """
    An associative dict of all genes

    Ensures that every gene is a single GeneSingleton instance
    and no duplicate instances are generated during parsing of the
    Gen-Pheno-HPO associations.

    This class is initilized once and genes are created by calling
    the instance of GeneDict to ensure that the same gene exists only once.

    For example ::

        Gene = GeneDict()
        gba = Gene(symbol='GBA')
        ezh2 = Gene(symbol='EZH2')
        gba_2 = Gene(symbol='GBA')

        gba is ezh2
        >> False
        gba is gba_2
        >> True

    Parameters
    ----------
    cols: list, default: ``None``
        Only used for backwards compatibility reasons.
        Should have the following entries

        * None
        * None
        * HGNC-ID
        * Gene symbol

    hgncid: int
        The HGNC ID
    symbol: str
        The gene symbol (alternative to name)

    Returns
    -------
    :class:`.GeneSingleton`
    """

    def __init__(self) -> None:
        self._indicies: Dict[int, GeneSingleton] = {}
        self._names: Dict[str, GeneSingleton] = {}

    def __call__(self, hgncid: int, symbol: str) -> GeneSingleton:
        try:
            return self._names[symbol]
        except KeyError:
            pass
        try:
            return self._indicies[hgncid]
        except KeyError:
            pass

        gene = GeneSingleton(id=hgncid, name=symbol)  # type: ignore

        self[gene] = gene
        self._indicies[hgncid] = gene
        self._names[symbol] = gene

        return gene

    def clear(self) -> None:
        """
        Removes all Genes to start with blank state.

        There are almot zero use-cases to ever call this method in
        client-side code. If you use code, that modifies the ``GeneDict``
        class a lot, you might use this. Under normal circumstences,
        this should not be needed.

        Currently, the primary use case is to clean up the state for
        unittests.
        """
        self._indicies.clear()
        self._names.clear()
        dict.clear(self)

    def get(self, query: Union[int, str], default: Any = None) -> GeneSingleton:
        """
        Allows client to query for a gene by both ID and symbol.
        This method is useful for client that do not want to add new
        genes

        Parameters
        ----------
        query: int or str
            The (most likely user supplied) query.
            Can be either the HGNC-ID or the gene symbol

        Returns
        -------
        GeneSingleton
            If a gene is found, it is returned. Otherwise an Error is raised
        """
        try:
            idx: int = int(query)
            return self._indicies[idx]
        except (ValueError, KeyError):
            idx = None  # type: ignore[assignment]  # desired

        try:
            return self._names[str(query)]
        except KeyError:
            raise KeyError("No gene found for query")


class DiseaseSingleton(Annotation):
    """
    This class represents a single disease.

    .. note::

        ``DiseaseSingleton`` should never be initiated directly,
        but only via the appropriate disease dictionary, e.g.
        :class:`.OmimDict` (:class:`.DiseaseDict`)
        to ensure that every disease is only created once.

    Attributes
    ----------
    id: int
        Disease ID

    name: str
        disease name

    hpo: set of :class:`pyhpo.term.HPOTerm`
        all HPOTerms associated to the disease
    negative_hpo: set of :class:`pyhpo.term.HPOTerm`
        HPOTerms not associated to the disease

    Parameters
    ----------
    id: int
        Disease ID
    name: str
        Disease name
    """

    diseasetype: str = "Undefined"
    negative_hpo: Set[int] = set()


class OmimDisease(DiseaseSingleton):
    diseasetype: str = "Omim"


class OrphaDisease(DiseaseSingleton):
    diseasetype: str = "Orpha"


class DecipherDisease(DiseaseSingleton):
    diseasetype: str = "Decipher"


class DiseaseDict(dict):
    """
    An associative dict of all Omim Diseases

    Ensures that every Omim Disease is a single OmimDisease instance
    and no duplicate instances are generated during parsing of the
    Gen-Pheno-HPO associations.

    This class is initilized once and diseases are created by calling
    the instance of ``DiseaseDict`` to ensure that the same disease
    exists only once.

    For example ::

        Disease = OmimDict()
        gaucher = Disease(diseaseid=1)
        fabry = Disease(diseaseid=2)
        gaucher_2 = Disease(diseaseid=1)

        gaucher is fabry
        >> False
        gaucher is gaucher_2
        >> True


    Parameters
    ----------
    cols: list, default: ``None``
        Only used for backwards compatibility reasons.
        Should have the following entries

        * None
        * Disease ID
        * Disease Name

    diseaseid: int
        The Disease ID
    name: str
        The disease name

    Returns
    -------
    :class:`.DiseaseSingleton`
    """

    disease_class: ClassVar = None

    def __init__(self) -> None:
        self._indicies: Dict[int, DiseaseSingleton] = {}

    def __call__(self, diseaseid: int, name: str) -> DiseaseSingleton:
        assert self.disease_class
        try:
            return self._indicies[diseaseid]
        except KeyError:
            pass

        disease: DiseaseSingleton = self.disease_class(id=diseaseid, name=name)

        self[disease] = disease
        self._indicies[diseaseid] = disease

        return disease

    def clear(self) -> None:
        """
        Removes all Diseases to start with blank state.

        There are almot zero use-cases to ever call this method in
        client-side code. If you use code, that modifies the ``DiseaseDict``
        class a lot, you might use this. Under normal circumstences,
        this should not be needed.

        Currently, the primary use case is to clean up the state for
        unittests.
        """
        self._indicies.clear()
        dict.clear(self)

    def get(self, query: Union[int, str], default: Any = None) -> DiseaseSingleton:
        """
        Allows client to query for a disease by ID.
        This method is useful for client that do not want to add new
        diseases

        Parameters
        ----------
        query: int
            The (most likely user supplied) query for Disease ID.

        Returns
        -------
        DiseaseSingleton
            If a disease is found, it is returned. Otherwise an Error is raised
        """
        try:
            idx = int(query)
            return self._indicies[idx]
        except ValueError:
            raise ValueError("Invalid Disease ID supplied")
        except KeyError:
            raise KeyError("No disease found for query")


class OmimDict(DiseaseDict):
    disease_class = OmimDisease


class OrphaDict(DiseaseDict):
    disease_class = OrphaDisease


class DecipherDict(DiseaseDict):
    disease_class = DecipherDisease


Omim = OmimDict()
Orpha = OrphaDict()
Decipher = DecipherDict()
Gene = GeneDict()



from operator import or_
from functools import reduce, lru_cache
from typing import Any, Dict, List, Optional, Set, Tuple

from pydantic import BaseModel, Field
from backports.cached_property import cached_property

from pyhpo.config import MODIFIER_IDS
from pyhpo.similarity import SimScore
from pyhpo.annotations import GeneSingleton
from pyhpo.annotations import OmimDisease, DecipherDisease, OrphaDisease
from pyhpo.parser.generics import id_from_string


class InformationContent(BaseModel):
    """
    InformationContent contains automatically calculated IC based on
    direct/indirect associations with genes, omim, orpha and decipher.
    IC instances are created automatically and accessed through
    :class:`pyhpo.term.HPOTerm` instances.

    Users can also register and calculate custom IC scores via
    :func:`pyhpo.term.InformationContent.set_custom`.
    """

    gene: float = 0.0  # Gene based IC
    omim: float = 0.0  # OMIM based IC
    orpha: float = 0.0  # OrphaNet based IC
    decipher: float = 0.0  # Decipher based IC
    custom: Dict[str, float] = Field(default_factory=dict)

    def __getitem__(self, key: str) -> float:
        """
        The IC is frequently accessed dynamically. e.g. in PyhPOAPI the
        kind of IC is specified in the query (omim / gene)
        Due to this, a dynamic access method is provided

        .. code-block:: python

            ic_kind = 'omim'
            term.information.content[ic_kind]

        """
        try:
            return float(self.__getattribute__(key))
        except AttributeError as err:
            if key in self.custom:
                return self.custom[key]
            else:
                raise AttributeError from err

    def set_custom(self, key: str, value: float) -> None:
        """
        Set the IC of a custom score

        Parameters
        ----------
        key: str
            The name of the information-content metric
        value: float
            The actual information content


        **Example:** ::

            for term in Ontology:
                # For some reason, you want to base the information content
                # on the depths of the Term in the ontology
                term.setcustom('depth',  term.shortest_path_to_root())

            # and now calculate similarity of two sets
            my_similarity = term_set_1.similarity(term_set_2, kind='depth')

        """
        self.custom[key] = value


class HPOTerm(BaseModel):
    """
    An HPOTerm instance can be build solely by itself,
    without knowledge of the actual Ontology. This is not recommended
    because it would miss all ontology features, such as parents, children,
    associated genes and diseaases etc.

    An HPOTerm instance should always be derived from the :class:`pyhpo.Ontology`
    """

    ###
    # Always present and mandatory
    ###

    id: str
    """
    The HPO identifier, e.g. ``HP:0000118``
    """

    name: str
    """
    The name of the HPO term, e.g. ``Abnormal axial skeleton morphology``
    """

    ###
    # Mandatory, calculated during initialization
    ###

    index: int
    """
    The integer representation of the HPO identifier
    """

    _hash: int

    ###
    # Mandatory for HPOTerm, but not always present in input
    ###

    comment: str = ""
    """
    The comment from the OBO source file
    """

    definition: str = ""
    """
    The definition from the OBO source file
    """

    _is_a: List[str] = []
    synonym: List[str] = []
    """
    A list of synonymous names for the term
    """

    xref: List[str] = []
    alt_id: List[str] = []

    ###
    # Special logic for some obsolete terms
    ###

    is_obsolete: bool = False
    replaced_by: Optional[str] = None
    consider: List[str] = []

    ###
    # Computed once all HPO Terms are present in the Ontology
    ###

    parents: Set["HPOTerm"] = set()
    """
    A set of all direct parent terms
    """

    children: Set["HPOTerm"] = set()
    """
    A set of all direct child terms
    """

    genes: Set[GeneSingleton] = set()
    """
    A set of all associated genes. Associated genes are inversely inherited from
    child terms as well
    """

    omim_diseases: Set[OmimDisease] = set()
    """
    A set of all associated Omim diseases. Associated diseases are inversely inherited from
    child terms as well
    """

    omim_excluded_diseases: Set[OmimDisease] = set()
    """
    A set of all explicitly non-associated Omim diseases. Non-associated diseases are inherited from
    parent terms as well
    """

    orpha_diseases: Set[OrphaDisease] = set()
    """
    A set of all associated Orpha diseases. Associated diseases are inversely inherited from
    child terms as well
    """

    orpha_excluded_diseases: Set[OrphaDisease] = set()
    """
    A set of all explicitly non-associated Orpha diseases.
    Non-associated diseases are inherited from parent terms as well
    """

    decipher_diseases: Set[DecipherDisease] = set()
    """
    A set of all associated Decipher diseases. Associated diseases are inversely inherited from
    child terms as well
    """

    decipher_excluded_diseases: Set[DecipherDisease] = set()
    """
    A set of all explicitly non-associated Decipher diseases.
    Non-associated diseases are inherited from parent terms as well
    """

    information_content: InformationContent = InformationContent()
    """
    The :class:`.InformationContent` of the HPO term.
    Multiple kinds of IC are automatically calculated,
    others can be manually calculated.
    """

    def __init__(self, **kwargs) -> None:  # type: ignore
        kwargs["index"] = id_from_string(kwargs["id"])
        super().__init__(**kwargs)
        self._hash = hash((self.index, self.name))
        self._is_a = kwargs.get("is_a", [])

    @cached_property
    def all_parents(self) -> Set["HPOTerm"]:
        hierarchy_set = reduce(or_, [set(path) for path in self.hierarchy])
        hierarchy_set.remove(self)
        return hierarchy_set

    @cached_property
    def hierarchy(self) -> Tuple[Tuple["HPOTerm", ...], ...]:
        """
        Calculates all paths from current term to Root term
        and returns each path as a Tuple of HPOTerms

        .. note::

            This function is expensive. To ensure better performance, the
            result is cached and all subsequent calls utilize the cache. Don't
            call ``hierarchy`` before the Ontology is fully built with all
            items.

        Returns
        -------
        tuple of tuple of :class:`.HPOTerm` s
            Tuple of paths. Each path is another tuple made up of HPOTerms
        """
        if not self.parents:
            return ((self,),)

        paths: List[Tuple["HPOTerm", ...]] = []
        for parent in self.parents:
            for path in parent.hierarchy:
                paths.append((self,) + path)

        return tuple(paths)

    @cached_property
    def is_modifier(self) -> bool:
        return int(self) in MODIFIER_IDS or bool(
            MODIFIER_IDS & {int(x) for x in self.all_parents}
        )

    def parent_ids(self) -> List[int]:
        return [id_from_string(item) for item in self._is_a]

    def parent_of(self, other: "HPOTerm") -> bool:
        """
        Checks if ``self`` is a direct or indirect parent of ``other``.

        Parameters
        ----------
        other: :class:`.HPOTerm`
            HPOTerm to check for lineage dependency

        Returns
        -------
        bool
            Is the HPOTerm a direct or indirect parent of another HPOTerms
        """
        return other.child_of(self)

    def child_of(self, other: "HPOTerm") -> bool:
        """
        Checks if ``self`` is a direct or indirect child of ``other``.

        Parameters
        ----------
        other: :class:`.HPOTerm`
            HPOTerm to check for lineage dependency

        Returns
        -------
        bool
            Is the HPOTerm a direct or indirect child of another HPOTerms
        """
        if self == other:
            raise RuntimeError("An HPO term cannot be parent/child of itself")

        return other in self.all_parents

    def common_ancestors(self, other: "HPOTerm") -> Set["HPOTerm"]:
        """
        Identifies all common ancestors
        of two HPO terms

        Parameters
        ----------
        other: :class:`.HPOTerm`
            Target HPO term for path finding

        Returns
        -------
        set
            Set of common ancestor HPOTerms
        """

        # Return the intersection of all ancestors of self and other.
        # Consider the following edge cases:
        # - self is in other.all_parents
        # - other is in self.all_parents
        # To account for these edge cases,
        # we first add self to self.all_parents
        # and other to other.all_parents
        self_ancestors: Set["HPOTerm"] = self.all_parents | set([self])
        other_ancestors: Set["HPOTerm"] = other.all_parents | set([other])
        return self_ancestors & other_ancestors

    def longest_path_to_root(self) -> int:
        """
        Calculates the longest path to root

        Returns
        -------
        int
            Maximum number of nodes until the root HPOTerm
        """
        return max([len(h) - 1 for h in self.hierarchy])

    def shortest_path_to_root(self) -> int:
        """
        Calculates the shortest path to root

        Returns
        -------
        int
            Minimum number of nodes until the root HPOTerm
        """
        return min([len(h) - 1 for h in self.hierarchy])

    def shortest_path_to_parent(
        self, other: "HPOTerm"
    ) -> Tuple[int, Tuple["HPOTerm", ...]]:
        """
        Calculates the shortest path to another HPO Term

        Parameters
        ----------
        other: HPOTerm
            parent HPOTerm instance

        Returns
        -------
        int
            Minimum number of nodes until the specified HPOTerm

            (float('inf') if ``other`` is not a parent.)
        tuple
            Tuple of all HPOTerm instances on the path

            (``None`` if ``other`` is not a parent)
        """
        if other not in self.all_parents and self != other:
            raise RuntimeError(f"{other.id} is not a parent of {self.id}")
        return_tuples: List[Tuple[int, Tuple["HPOTerm", ...]]] = []
        for path in self.hierarchy:
            try:
                i = path.index(other)
                return_tuples.append((i, path[: i + 1]))
            except ValueError:
                pass

        try:
            return sorted(return_tuples, key=lambda x: x[0])[0]
        except IndexError as err:
            raise RuntimeError(
                f"Unable to determine path to parent term {other.name}"
            ) from err

    def longest_path_to_bottom(self, level: int = 0) -> int:
        """
        Calculates how far the most distant child is apart

        Parameters
        ----------
        level: int
            Offset level to indicate for calculation
            Default: 0

        Returns
        -------
        int:
            Number of steps to most distant child

        """
        if len(self.children):
            return max(
                [child.longest_path_to_bottom(level + 1) for child in self.children]
            )
        else:
            return level

    def path_to_other(
        self, other: "HPOTerm"
    ) -> Tuple[int, Tuple["HPOTerm", ...], int, int]:
        """
        Identifies the shortest connection between
        two HPO terms

        Parameters
        ----------
        other: HPOTerm
            Target HPO term for path finding

        Returns
        -------
        int
            Length of path
        tuple
            Tuple of HPOTerms in the path
        int
            Number of steps from term-1 to the common parent
        int
            Number of steps from term-2 to the common parent

        """
        common = self.common_ancestors(other)

        paths = []
        for term in common:
            path1 = self.shortest_path_to_parent(term)
            path2 = other.shortest_path_to_parent(term)

            total_path = path1[1] + tuple(reversed(path2[1]))[1:]
            paths.append(
                (int(path1[0] + path2[0]), total_path, int(path1[0]), int(path2[0]))
            )
        return sorted(paths, key=lambda x: x[0])[0]

    def count_parents(self) -> int:
        """
        Calculates total number of ancestral HPO Terms

        Returns
        -------
        int
            The number of all ancestral HPO Terms
        """
        return sum([parent.count_parents() + 1 for parent in self.parents])

    def similarity_score(
        self, other: "HPOTerm", kind: Optional[str] = None, method: Optional[str] = None
    ) -> float:
        """
        Calculate the similarity between this and another HPO-Term
        It uses :class:`pyhpo.similarity.base._Similarity` underneath

        Parameters
        ----------
        other: `HPOTerm`
            Other HPO term to compare similarity to

        kind: str, default ``''``
            Which kind of information content should be calculated.
            Default option is defined in `pyhpo.similarity.base._Similarity`

            Available options:

            * **omim**
            * **orpha**
            * **decipher**
            * **gene**

        method: string, default ``''``
            The method to use to calculate the similarity.
            Default option is defined in `pyhpo.similarity.base._Similarity`

            Available options:

            * **resnik** - Resnik P, Proceedings of the 14th IJCAI, (1995)
            * **lin** - Lin D, Proceedings of the 15th ICML, (1998)
            * **jc** - Jiang J, Conrath D, ROCLING X, (1997)
              Implementation according to R source code
            * **jc2** - Jiang J, Conrath D, ROCLING X, (1997)
              Implementation according to paper from R ``hposim`` library
              Deng Y, et. al., PLoS One, (2015)
            * **rel** - Relevance measure - Schlicker A, et.al.,
              BMC Bioinformatics, (2006)
            * **ic** - Information coefficient - Li B, et. al., arXiv, (2010)
            * **graphic** - Graph based Information coefficient -
              Deng Y, et. al., PLoS One, (2015)
            * **dist** - Distance between terms
            * Additional methods can be registered separately (
              see :class::`pyhpo.similarity.base._Similarity`)

        Raises
        ------
        RuntimeError
            The specified ``method`` does not exist
        NotImplementedError
            This error can only occur with custom Similarity-Score
            methods that do not have a ``similarity`` method defined.
        AttributeError
            The information content for ``kind`` does not exist
        """
        return SimScore(self, other, kind, method)

    @lru_cache(maxsize=128)
    def cached_similarity_score(
        self, other: "HPOTerm", kind: str = "", method: str = ""
    ) -> float:
        """
        This is a LRU-chached alias of
        :func:`pyhpo.term.HPOTerm.similarity_score`
        """
        return self.similarity_score(other, kind, method)

    def toJSON(self, verbose: bool = False) -> dict:
        """
        Creates a JSON-like object of the HPOTerm

        Parameters
        ----------
        verbose: bool, default ``False``
            Include extra properties

        Returns
        -------
        dict
            A dictionary with the main properties of the HPOTerm


        **Example:** ::

            >>> terms[2].toJSON()
            {
                'name': 'Abnormality of body height',
                'id': 'HP:0000002',
                'int': 2
            }

            >>> terms[2].toJSON(verbose=True)
            {
                'name': 'Abnormality of body height',
                'synonym': ['Abnormality of body height'],
                'comment': None,
                'def': '"Deviation from the norm of height with respect [...]',
                'xref': ['UMLS:C4025901'],
                'is_a': ['HP:0001507 ! Growth abnormality'],
                'id': 'HP:0000002',
                'int': 2
            }

        """
        res = {"int": int(self), "id": self.id, "name": self.name}

        if verbose:
            res["definition"] = self.definition
            res["comment"] = self.comment
            res["synonym"] = self.synonym
            res["xref"] = self.xref
            res["is_a"] = self._is_a
            res["ic"] = self.information_content.model_dump()

        return res

    def to_obo(self) -> str:
        raise NotImplementedError("Method is missing")

    def __hash__(self) -> int:
        """
        The hash is precalcuated during initialization
        """
        return self._hash

    def __int__(self) -> int:
        return self.index

    def __eq__(self, t2: Any) -> bool:
        return hash(self) == hash(t2) and isinstance(t2, HPOTerm)

    def __lt__(self, other: Any) -> bool:
        return int(self) < int(other)

    def __str__(self) -> str:
        return "{} | {}".format(self.id, self.name)

    def __repr__(self) -> str:
        return f"HPOTerm(id='{self.id}', name='{self.name}', " f"is_a={self._is_a})"

    class Config:
        arbitrary_types_allowed = True
        ignored_types = (cached_property,)



"""
Parse the OBO flat-file
"""

import os
from typing import Callable, Dict, Iterator, List

from pyhpo.config import TRUTH


FILENAME = "hp.obo"


class Metadata:
    format_version: str
    data_version: str
    header: List[str] = []

    @classmethod
    def add_header_row(cls, row: str) -> None:
        cls.header.append(row)


class Converter:
    key_conversion: Dict[str, str] = {"def": "definition"}

    type_conversions: Dict[str, Callable] = {}

    @classmethod
    def add_type_conversion(cls, key: str, func: Callable) -> None:
        cls.type_conversions[key] = func

    @staticmethod
    def array_to_str(value: List[str], key: str, values: List[List[str]]) -> str:
        if len(value):
            return value[0]
        return ""

    @staticmethod
    def array_to_bool(value: List[str], key: str, values: List[List[str]]) -> bool:
        if not len(value):
            return False
        return value[0].lower() in TRUTH

    @staticmethod
    def parse_synonym(value: List[str], key: str, values: List[List[str]]) -> List[str]:
        """
        Extracts the synonym from the synonym data line in the obo file format

        Parameters
        ----------
        synonym: str
            value part of synonym-data line of obo file

            e.g: "Multicystic dysplastic kidney" EXACT []

        Returns
        -------
        str
            Actual synonym title

            e.g.: Multicystic dysplastic kidney
        """
        return [x.split('"')[1] for x in value]


Converter.add_type_conversion("id", Converter.array_to_str)
Converter.add_type_conversion("name", Converter.array_to_str)
Converter.add_type_conversion("comment", Converter.array_to_str)
Converter.add_type_conversion("definition", Converter.array_to_str)
Converter.add_type_conversion("is_obsolete", Converter.array_to_bool)
Converter.add_type_conversion("replaced_by", Converter.array_to_str)
Converter.add_type_conversion("synonym", Converter.parse_synonym)


def terms_from_file(data_folder: str) -> Iterator[dict]:
    """
    Reads an obo file line by line to yield
    a dict for building an HPOTerm

    Parameters
    ----------
    data_folder:
        Full path to ``obo`` file

    """
    filename = os.path.join(data_folder, FILENAME)

    with open(filename) as fh:
        # everything above the first [Term] is header
        # and thus must not be parsed as term
        for line in fh:
            line = line.strip()
            if line == "[Term]":
                break
            else:
                Metadata.add_header_row(line)

        term_section: List[str] = []
        for line in fh:
            line = line.strip()
            if line == "[Term]":
                yield parse_obo_section(term_section)
                term_section = []
            elif line == "[Typedef]":
                # we're currently not parsing an Typedef section.
                # Since they only appear at the end of the OBO file
                # we're stopping the parsing here.
                # TODO: Instead of break, add logic to skip all Typedef
                # sections and continue with term parsing
                break
            else:
                term_section.append(line)

        yield parse_obo_section(term_section)


def parse_obo_section(term_section: List[str]) -> dict:
    """
    Parses the section of an OBO file for one single HPO term

    Parameters
    ----------
    term_section:
        Lines of the ``obo`` file that describe the HPO term
    """
    term_data = {}
    for line in term_section:
        if line == "":
            continue
        key, value = line.split(":", 1)
        if key not in term_data:
            term_data[key] = [value.strip()]
        else:
            term_data[key].append(value.strip())
    term_dict = _convert_dict_keys(term_data)
    term_dict = _convert_value_types(term_data)
    return term_dict


def _convert_dict_keys(term_data: dict) -> dict:
    """
    The HPO obo flat file contains some unfortunate attribute names.
    This function will convert them into the actual attributes
    for ``HPOTerm``
    """

    for old, new in Converter.key_conversion.items():
        term_data[new] = term_data.pop(old, [])

    return term_data


def _convert_value_types(term_data: dict) -> dict:
    for key, convert in Converter.type_conversions.items():
        term_data[key] = convert(term_data.get(key, []), key, term_data)

    return term_data



from typing import Iterable, Any, Optional, List, Union, Tuple


class Matrix:
    """
    # noqa: E501

    Poor man's implementation of a DataFrame/Matrix

    This is used to calculate similarities between HPO sets
    and is surprisingly much faster than using pandas DataFrames

    .. note::

        Pandas::

            ===== COMPARING SETS ======
            23806489 function calls (23770661 primitive calls) in 19.705 seconds

            ncalls  tottime  percall  cumtime  percall filename:lineno(function)
            ....
            9900    0.267    0.000   19.106    0.002 set.py:318(similarity)
            9900    1.124    0.000   14.330    0.001 set.py:477(_sim_score)
            ....

        Matrix::

            ===== COMPARING SETS ======
            12870433 function calls in 6.642 seconds

            ncalls  tottime  percall  cumtime  percall filename:lineno(function)
            ....
            9900    0.048    0.000    6.424    0.001 set.py:316(similarity)
            9900    0.928    0.000    5.112    0.001 set.py:432(_sim_score)
            ....

    .. warning::

        This `Matrix` should not be used as a public interface.
        It's only used internally for calculations.

    Parameters
    ----------
    rows: int
        The number of rows in the Matrix
    cols: int
        The number of columns in the Matrix
    data: list of values, default ``None``
        A list with values to fill the Matrix.

    Attributes
    ----------
    n_rows: int
        The number of rows in the Matrix

    n_cols: int
        The number of columns in the Matrix

    rows: iterator
        Iterator over all rows

        **Example:** ::

            print(matrix)

            >>    ||   0|   1|   2|   3|
            >> =========================
            >> 0  ||  11|  12|  13|  14|
            >> 1  ||  21|  22|  23|  24|
            >> 2  ||  31|  32|  33|  34|

            for row in matrix.rows:
                print(row)

            >> [11, 12, 13, 14]
            >> [21, 22, 23, 24]
            >> [31, 32, 33, 34]

    columns: iterator
        Iterator over all columns

        **Example:** ::

            print(matrix)

            >>    ||   0|   1|   2|   3|
            >> =========================
            >> 0  ||  11|  12|  13|  14|
            >> 1  ||  21|  22|  23|  24|
            >> 2  ||  31|  32|  33|  34|

            for col in matrix.columns:
                print(col)

            >> [11, 21, 31]
            >> [12, 22, 32]
            >> [13, 23, 33]
            >> [14, 24, 34]

    """

    def __init__(self, rows: int, cols: int, data: Optional[List[Any]] = None):
        self.n_rows = rows
        self.n_cols = cols
        self._data: List[Any]
        if data is None:
            self._data = [None] * rows * cols
        elif len(data) == rows * cols:
            self._data = data
        else:
            raise RuntimeError("Wrong number of data items in `data`")

    def __getitem__(self, key: Tuple[Optional[int], Optional[int]]) -> Any:
        idx = self._get_key_indicies(key)
        return self._data[idx]

    def __setitem__(self, key: Tuple[Optional[int], Optional[int]], val: Any) -> None:
        idx = self._get_key_indicies(key)

        if isinstance(idx, int) or len(self._data[idx]) == len(val):
            self._data[idx] = val
        else:
            raise ValueError("Different length of matrix subset and values")

    def _get_key_indicies(
        self, key: Tuple[Optional[int], Optional[int]]
    ) -> Union[int, slice]:
        row = key[0]
        col = key[1]

        if row is None and isinstance(col, int):
            # Return one column
            return slice(col, self.n_rows * (self.n_cols) + col, self.n_cols)

        if col is None and isinstance(row, int):
            # Return one row
            return slice(row * self.n_cols, row * self.n_cols + self.n_cols, 1)

        if isinstance(row, int) and isinstance(col, int):
            if row > self.n_rows - 1:
                raise RuntimeError("Invalid row number: {}".format(row))
            if col > self.n_cols - 1:
                raise RuntimeError("Invalid column number: {}".format(col))
            return row * self.n_cols + col

        raise RuntimeError("Invalid arguments for Matrix subset")

    @property
    def rows(self) -> Iterable[Any]:
        for x in range(self.n_rows):
            yield self[x, None]

    @property
    def columns(self) -> Iterable[Any]:
        for x in range(self.n_cols):
            yield self[None, x]

    def __str__(self) -> str:
        maxlength = max([len(str(self.n_cols))] + [len(str(x)) for x in self._data]) + 2
        idxlength = len(str(self.n_rows)) + 2

        s = ""

        s += "{}||".format("".rjust(idxlength))
        s += "".join(
            ["{}|".format(str(x).rjust(maxlength)) for x in range(self.n_cols)]
        )
        s += "\n" + "=" * len(s)

        for idx, item in enumerate(self._data):
            if idx % self.n_cols == 0:
                s += "\n{}||".format(str(int(idx / self.n_cols)).ljust(idxlength))
            s += "{}|".format(str(item).rjust(maxlength))

        return s



from typing import Dict, List, Optional, Type

from pydantic import BaseModel

import pyhpo


class _Similarity(BaseModel):
    dispatch: Dict[str, "SimilarityBase"] = {}
    kind: str = "omim"
    method: str = "graphic"

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: Optional[str] = "",
        method: Optional[str] = "",
    ) -> float:
        kind = kind or self.kind
        method = method or self.method
        try:
            similarity = self.dispatch[method]
        except KeyError as err:
            raise RuntimeError(
                f"Unknown method {method} to calculate similarity"
            ) from err

        dependencies: List[float] = [
            self(term1, term2, kind, dep) for dep in similarity.dependencies
        ]

        return similarity(term1, term2, kind, dependencies)

    def register(self, name: str, similarity_class: Type["SimilarityBase"]) -> None:
        self.dispatch[name] = similarity_class()


class SimilarityBase(BaseModel):
    """
    Base class to use for custom similarity calculations.

    Custom implementation must inherit from
    :class:`pyhpo.similarity.base.SimilarityBase` and provide a
    ``__call__`` method with the same signature as
    :func:`pyhpo.similarity.base.SimilarityBase.__call__`

    You can also provide a list of ``dependencies`` of other
    similarity methods that should be called beforehand. Results
    of these calls will be passed as ``dependencies`` parameter to
    the ``__call__`` method.
    """

    dependencies: List[str] = []

    def __call__(
        self,
        term1: "pyhpo.HPOTerm",
        term2: "pyhpo.HPOTerm",
        kind: str,
        dependencies: List[float],
    ) -> float:
        """
        This method does the actual calculation of the similarity.
        This method must be provided in all custom similarity classes

        Parameters
        ----------
        term1:
            One of the two terms to compare
        term2:
            The other of the two terms to compare
        kind:
            This can be an extra parameter, ususally ``omim`` or ``gene``
            to specify which annotations to consider for similarity
        depdencies:
            A list of other calculation-results that should be calculated
            beforehand. This is not needed at all, but helpful if your
            implementation builds on an already existing similarity
            calculation.
        """
        raise NotImplementedError("A similarity class requires a __call__ function")


SimScore = _Similarity()



import os
import urllib.request
import logging
from typing import Optional

from pyhpo.parser.obo import FILENAME as hpo_file
from pyhpo.parser.genes import FILENAME as gene_file
from pyhpo.parser.diseases import FILENAME as pheno_file


"""
Downloads the required data for HPO Ontology and the annotations
"""
logger = logging.getLogger(__name__)


URLS = {
    "HPO_ONTOLOGY": "http://purl.obolibrary.org/obo/hp.obo",
    "HPO_GENE": "http://purl.obolibrary.org/obo/hp/hpoa/phenotype_to_genes.txt",  # noqa: E501
    "HPO_PHENO": "http://purl.obolibrary.org/obo/hp/hpoa/phenotype.hpoa",
}

FILENAMES = {"HPO_ONTOLOGY": hpo_file, "HPO_GENE": gene_file, "HPO_PHENO": pheno_file}


def make_backup(filename: str) -> None:
    logger.debug("Backup not yet implemented")


def download_data(data_dir: Optional[str] = None) -> None:
    if data_dir is None:
        data_dir = os.path.realpath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")
        )
    else:
        data_dir = os.path.realpath(data_dir)

    logger.debug("Downloading data to %s", data_dir)

    for url in URLS:
        logger.debug("Downloading %s", url)
        filename = os.path.join(data_dir, FILENAMES[url])
        if os.path.exists(filename):
            logger.warning("%s exists already. Backing up old data", filename)
            make_backup(filename)
        urllib.request.urlretrieve(URLS[url], filename)
